/* 
   IBM EUI interface, using the EUI-level 1 routines

   UNDER CONSTRUCTION

   Pending changes:
   ASYNCRecvId_t should be a pointer to a structure that holds the type and
   source (since these values are returned asynchronously).  A simple heap of
   such structures can be allocated; by using the pointer, they will work 
   even for Fortran users.

   We may need to convert from a tag of -1 to DONTCARE

   We may need to call mpc_task_query to get ALLGRP and DONTCARE

   We may need our routines for I_VADD etc.
 */

#ifndef __commeui
#define __commeui

/* #include <mp_eui.h> */
/* Note that mp_eui doesn't check to see if it has already been included... */
/* Note also that mp_eui.h is INCOMPATIBLE with POSIX_SOURCE (needed by PETSc).
   Thus, we include */
#include <sys/types.h>
typedef ulong_t ulong;

/* EUI-H is Fortran ONLY, and has no include file */
/* There IS no EUI include file now. (The documentation is inconsistent about
   this) */
#ifndef euih
#define DONTCARE -1
/* #include <mp_eui.h> */
#else
#define DONTCARE -1
#endif

#ifndef __commh
#include "comm/comm.h"
#endif

extern int __EUILEN, __EUIFROM, __EUITYPE;
extern int __NUMNODES, __MYPROCID;
extern int __ALLGRP;

extern void PIEUI_BSend();

#define MSG_INT   0
#define MSG_LNG   1
#define MSG_FLT   2
#define MSG_DBL   3
#define MSG_OTHER 4
#define MSG_SHRT  5

/*
   The C interface seems to call the Fortran interface so why not use
   the Fortran interface directly. 
*/
/* 
   EUI-H, an experimental, high-performance version of EUI, only has the 
   Fortran interface.  In order to allow the same object libraries to be used
   for both EUI and EUI-H, we have chosen the Fortran interface.  This 
   may involve slightly more overhead, since some parameters must be passed
   as addresses rather than values.
 */
#define USE_EUI_FORTRAN_INTERFACE 

/* There are native global ops; I just haven't included them yet */
#define NO_NATIVE_GLOBAL
#define NO_FORCE

/* Notes on these routines:
   The recv routines return information on the type, len, and source
   in the arguments used to pass those values as selectors.  Thus
   we need to copy those values to temps before calling the receive routines
 */   
#if defined(USE_EUI_FORTRAN_INTERFACE)
/*	     mp_bsend(buffer,&__l,&__t,&__ty);\ */
#define SENDSYNC(type,buffer,length,to,datatype) \
         { int __l = length, __t = to, __ty = type; \
         LOGBSENDSTART(length,type,to);\
         PIEUI_BSend( type, buffer, length, to, datatype );\
         LOGBSENDEND(type);}
#else
	     /*mpc_bsend(buffer,length,to,type);\ */
#define SENDSYNC(type,buffer,length,to,datatype) \
        {LOGBSENDSTART(length,type,to);\
	     PIEUI_BSend( type, buffer, len, to, datatype );\
         LOGBSENDEND(type);}
#endif
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define RECVSYNC(type,buffer,length,datatype) {int __l = length; \
        __EUIFROM=DONTCARE;\
        __EUITYPE=type; LOGBRECVSTART(length,type);\
        mp_brecv(buffer,&__l,&__EUIFROM,&__EUITYPE,&__EUILEN);\
        LOGBRECVEND(type,__EUILEN,__EUIFROM);}
#else
#define RECVSYNC(type,buffer,length,datatype) {__EUIFROM=DONTCARE;\
        __EUITYPE=type;; LOGBRECVSTART(length,type);\
        mpc_brecv(buffer,length,&__EUIFROM,&__EUITYPE,&__EUILEN);\
        LOGBRECVEND(type,__EUILEN,__EUIFROM);}
#endif
#define SENDSYNCNOMEM(type,buffer,length,to,datatype) \
        SENDSYNC(type,buffer,length,to,datatype)
#define RECVSYNCNOMEM(type,buffer,length,datatype)  \
        RECVSYNC(type,buffer,length,datatype);
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define RECVSYNCUNSZ(type,bufp,size,datatype) {\
        do {__EUIFROM=DONTCARE;__EUITYPE=type;\
	    mp_probe(&__EUIFROM,&__EUITYPE,&__EUILEN);}while(__EUILEN<0);\
	size=__EUILEN;\
        bufp=(char*)MALLOC(__EUILEN);\
        RECVSYNC(type,bufp,__EUILEN,datatype);}
#else
#define RECVSYNCUNSZ(type,bufp,size,datatype) {\
        do {__EUIFROM=DONTCARE;__EUITYPE=type;\
	    mpc_probe(&__EUIFROM,&__EUITYPE,&__EUILEN);}while(__EUILEN<0);\
	size=__EUILEN;\
        bufp=(char*)MALLOC(__EUILEN);\
        RECVSYNC(type,bufp,__EUILEN,datatype);}
#endif

typedef int ASYNCRecvId_t;
/* Eventually, need:
typedef struct { int recvlen, id; } ASYNCRecvId_t;
 */
typedef int ASYNCSendId_t;
/* Asynchronous routines */
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define SENDASYNC(type,buffer,length,to,datatype,id) \
        {int __ty = type, __l = length, __t = to; \
        LOGNSENDSTART(length,type,to);\
	     mp_send(buffer,&__l,&__t,&__ty,&(id));\
		 LOGNSENDEND(type);}
#else
#define SENDASYNC(type,buffer,length,to,datatype,id) \
        {LOGNSENDSTART(length,type,to);\
	     mpc_send(buffer,length,to,type,&(id));\
		 LOGNSENDEND(type);}
#endif
/* Design problem:  __EUIFROM will be set when the message comes in.
   Thus we really need to modify AsyncRecvId_t to be a structure
   containing the source field! */
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define RECVASYNC(type,buffer,length,datatype,id) \
     {int __l = length; __EUIFROM=DONTCARE;__EUITYPE=type;\
     LOGNRECVSTART(length,type);\
     mp_recv(buffer,&__l,&__EUIFROM,&__EUITYPE,&(id));LOGNRECVEND(type);}
#else
#define RECVASYNC(type,buffer,length,datatype,id) \
     {__EUIFROM=DONTCARE;__EUITYPE=type;\
     LOGNRECVSTART(length,type);\
     mpc_recv(buffer,length,&__EUIFROM,&__EUITYPE,&(id));LOGNRECVEND(type);}
#endif
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define SENDWAIT(type,buffer,length,to,datatype,id) {\
    int d; LOGWSENDSTART(length,type,to);mp_wait(&(id),&d);LOGWSENDEND(type);}
#else
#define SENDWAIT(type,buffer,length,to,datatype,id) {\
    int d; LOGWSENDSTART(length,type,to);mpc_wait(&(id),&d);LOGWSENDEND(type);}
#endif
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define RECVWAIT(type,buffer,length,datatype,id)     \
        {LOGWRECVSTART(type);mp_wait(&(id),&__EUILEN);\
	     LOGWRECVEND(type,__EUILEN,-1);}
#else
#define RECVWAIT(type,buffer,length,datatype,id)     \
        {LOGWRECVSTART(type);mpc_wait(&(id),&__EUILEN);\
	     LOGWRECVEND(type,__EUILEN,-1);}
#endif
#define SENDASYNCNOMEM(type,buffer,length,to,datatype,id) \
        SENDASYNC(type,buffer,length,to,datatype,id)
#define RECVASYNCNOMEM(type,buffer,length,datatype,id) \
        RECVASYNC(type,buffer,length,dataype,id)
#define SENDWAITNOMEM(type,buffer,length,to,datatype,id)  \
        SENDWAIT(type,buffer,length,to,datatype,id)  
#define RECVWAITNOMEM(type,buffer,length,datatype,id)     \
        RECVWAIT(type,buffer,length,datatype,id)
/* does this have Fortran version?? */
#define ASYNCPROBE(type) mppi_asyncprobe(type)
/* Status in EUI returns the number of bytes; 0 is a valid length.  */
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define ASYNCDONE(id)    ( mp_status(&(id)) > -1 )
#else
#define ASYNCDONE(id)    ( mpc_status((id)) >= 0 )
#endif

/* Note: mpc_cancel missing from June 28 draft of EUI */
#ifdef NEWER
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define RECVCANCEL(id)   mp_cancel(&(id))
#define SENDCANCEL(id)   mp_cancel(&(id))
#else
#define RECVCANCEL(id)   mpc_cancel(id)
#define SENDCANCEL(id)   mpc_cancel(id)
#endif

#else
/* Cancel can fail, so this isn't too bad a fallback */
#define RECVCANCEL(id)   
#define SENDCANCEL(id)   
#endif

/* No Forcetype routines.  */
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define SYNCPROBE(type) {int __t = type; do {\
   __EUIFROM=DONTCARE;mp_probe(&__EUIFROM,&__t,&__EUILEN);}while(__EUILEN<0);}
#else
#define SYNCPROBE(type) {int __t = type do {\
   __EUIFROM=DONTCARE;mpc_probe(&__EUIFROM,&__t,&__EUILEN);}while(__EUILEN<0);}
#endif
#define RECVPROBED(type,buffer,length,datatype) \
        RECVSYNC(type,buffer,length,datatype)

#define RECVLEN()  __EUILEN
#define RECVFROM() __EUIFROM
#define RECVTYPE() __EUITYPE

#define MSGALLOCSEND(msg,maxmsg,type)  msg = (type *)MALLOC(maxmsg)
#define MSGFREESEND(msg)               FREE(msg)
#define MSGALLOCRECV(msg,maxmsg,type)  msg = (type *)MALLOC(maxmsg)
#define MSGFREERECV(msg)               FREE(msg)

/* 
   There are global operations, even on subsets!
   This will require some support by procset, since the procset code
   assumed that no-one would provide this functionality.  Procset
   should provide hooks for this, and there should be a procset 
   implementation layer.

   Problems:
   MPC_REDUCE is a "half reduction"; like the PICL versions.
   MPC_COMBINE is the full reduction.
   
   For now, we'll leave support for these operations out.
   
 */
#if defined(EUI_GLOBAL_AVAILABLE)
extern void i_vadd(), d_vadd(), f_vadd(), i_vmax(), d_vmax(), i_vmin(),
            d_vmin(), i_vor();
#if defined(USE_EUI_FORTRAN_INTERFACE)
/* #define GSCATTERGLOB(buf,size,issrc,datatype) */
#define GSCATTERSRCGLOB(buf,size,src,datatype) \
    {int _size=size,_src=src;mp_bcast( buf, &_size, &_src, &__ALLGRP );}
#define GSYNCGLOB() mp_sync(&__ALLGRP)
#define GISUMGLOB(val,n,work) \
    {int _size = n*sizeof(int); \
     mp_combine(val,val,i_vadd,&_size,&__ALLGRP);}
#define GDSUMGLOB(val,n,work) \
    {int _size=n*sizeof(double);\
	 mp_combine(val,val,&_size,d_vadd,&__ALLGRP);}
#define GFSUMGLOB(val,n,work) \
    {int _size=n*sizeof(float); \
     mp_combine(val,val,&_size,f_vadd,&__ALLGRP);}
#define GIMAXGLOB(val,n,work) \
    {int _size = n*sizeof(int);\
	 mp_combine(val,val,&_size,i_vmax,&__ALLGRP);}
#define GDMAXGLOB(val,n,work) \
    { int _size = n *sizeof(double);\
    mp_combine(val,val,&_size,d_vmax,&__ALLGRP);}
#define GIMINGLOB(val,n,work) \
    {  int _size = n * sizeof(int);\
    mp_combine(val,val,&_size,i_vmin,&__ALLGRP);}
#define GDMINGLOB(val,n,work) \
    { int _size = n *sizeof(double);\
    mp_combine(val,val,&_size,d_vmin,&__ALLGRP);}
#define GIORGLOB(val,n,work) \
    { int _size = n * sizeof(int); \
    mp_combine(val,val,&_size,i_vor,&__ALLGRP);}

#else
/* #define GSCATTERGLOB(buf,size,issrc,datatype) */
#define GSCATTERSRCGLOB(buf,size,src,datatype) \
    mpc_bcast( buf, size, src, __ALLGRP )
#define GSYNCGLOB() mpc_sync(__ALLGRP)
#define GISUMGLOB(val,n,work) \
    mpc_combine(val,val,n*sizeof(int),i_vadd,__ALLGRP)
#define GDSUMGLOB(val,n,work) \
    mpc_combine(val,val,n*sizeof(double),d_vadd,__ALLGRP)
#define GFSUMGLOB(val,n,work) \
    mpc_combine(val,val,n*sizeof(float),f_vadd,__ALLGRP)
#define GIMAXGLOB(val,n,work) \
    mpc_combine(val,val,n*sizeof(int),i_vmax,__ALLGRP)
#define GDMAXGLOB(val,n,work) \
    mpc_combine(val,n*sizeof(double),d_vmax,__ALLGRP)
#define GIMINGLOB(val,n,work) \
    mpc_combine(val,val,n*sizeof(int),i_vmin,__ALLGRP)
#define GDMINGLOB(val,n,work) \
    mpc_combine(val,val,n*sizeof(double),d_vmin,__ALLGRP)
#define GIORGLOB(val,n,work) \
    mpc_combine(val,val,n*sizeof(int),i_vor,__ALLGRP)
#endif

#endif

/* NUMNODES and MYPROCID need to be set by the init routine */
#define NUMNODES __NUMNODES
#define MYPROCID __MYPROCID
#if defined(USE_EUI_FORTRAN_INTERFACE)
#define MSGTYPERANGE(low,high) {int nbuf[2],two=2;\
    mp_task_query(nbuf,&two,&two);\
    *(low)=nbuf[0];*(high)=nbuf[1];}
#else
#define MSGTYPERANGE(low,high) {int nbuf[2];mpc_task_query(nbuf,2,2);\
				    *(low)=nbuf[0];*(high)=nbuf[1];}
#endif
#if defined(euih)
#define MSGSIZES(min,max)      {*min=0;*max=4096;}
#else
#define MSGSIZES(min,max)      {*min=0;*max=128;}
#endif
#define MSGDISTANCE(from,to)   (from != to)
#define MSGDIAMETER            (1)

#endif


