#ifndef lint
static char SCCSid[] = "@(#) ./comm/io/pieach.c 12/01/93";
#endif

/* This file contains routines for having each (all) tasks open a file
   and prepare to write to it.  This provides a simple parallel IO mode.

   A special situation involves managing the default file name.  The
   expectation is that the "local" files are identified by some unique
   prefix.  See PIiOpenEach below.

 */

#include <stdio.h>
#include <ctype.h>
#if defined(cray)
#include <errno.h>
#else
extern int errno;
#endif
#include <sys/errno.h>
#include "tools.h"
#include "comm/comm.h"
#include "comm/io/pio.h"
#ifndef SEEK_SET
#include <unistd.h>
#endif

#define READ  (*fp->read)
#define WRITE (*fp->write)
#define LSEEK (*fp->lseek)

void PIiConvertName();

#ifndef PIF_FILEPREFIXDEF

#if defined(intelnx)
/* This requires the intel version to use cread/cwrite */
#define PIF_FILEPREFIXDEF "/cfs/"
#else
#define PIF_FILEPREFIXDEF "/tmp/%u/"
#endif

#endif

/*+
   PIiOpenEach - Open a parallel file 

   The name PIF_FILEPREFIX is an environment variable that may be used to
   select the file prefix.  By default, it is /tmp/%u/.  Intel systems may
   be compiled with a different default (define PIF_FILEPREFIXDEF). 
   %u is the user's login name and %n is the node number. 
+*/
int PIiOpenEach( name, procset, mode, pmode, fmode )
char    *name;
ProcSet *procset;
int     mode, pmode, fmode;
{
int  fd, errcnt, work;
char nameout[1024];

/* Get the modified name */
PIiConvertName( name, nameout );

/* Create directories as necessary */
SYMakeAllDirs( nameout, fmode );

/* Open the file */
fd = open( nameout, mode, fmode );

/* Must let all processors know if the file was successfully opened, 
   and how to identify a message for the file */

/* Check that everyone succeeded. */
errcnt = fd < 0;
PIgisum( &errcnt, 1, &work, procset );

if (errcnt) {
    if (fd) close( fd );
    fd = 0;
    fprintf( stderr, "Failed to open file %s (%s)\n", name, nameout );
    }

return fd;
}

void PIiConvertName( name, nameout )
char *name, *nameout;
{
char *pin, *pout, *pu;
char uname[20];
char *fileprefix;
char namein[1024];
extern char *getenv();

/* Get the prefix */
fileprefix = getenv( "PIF_FILEPREFIX" );
if (!fileprefix) 
    strcpy( namein, PIF_FILEPREFIXDEF );
else 
    strcpy( namein, fileprefix );
if (namein[strlen(namein) - 1] != '/')
    strcat( namein, "/" );
strcat( namein, name );

pin  = namein;
pout = nameout;
/* Look for % */
while (*pin) {
    if (*pin == '%') {
	if (pin[1] == 'u') {
	    SYGetUserName( uname, 20 );
	    pu = uname;
	    while (*pu) *pout++ = *pu++;
	    }
	else if (pin[1] == 'n') {
	    sprintf( pout, "%d", PImytid );
	    while (*pout) pout++;
	    }
	pin += 2;
	}
    else
	*pout++ = *pin++;
    }
/* Don't forget to add the terminating character */
*pout = 0;
}


/* 
   Routines for "distributed" read and write.  These are just pass-throughs
   to the read/write routines 
 */
void PIiPwriter( fp, buf, n, offset )
PIFILE *fp;
char   *buf;
int    n, offset;
{
int err;

/* printf( "lWriting %d bytes at offset %d\n", n-sizeof(long), offset ); */
/* LSEEK( fp, offset, SEEK_SET, 1 ); */
err = WRITE( fp, buf+sizeof(long), n );
if (err != n) {
    SETERRC(1,"Write failed" );
    }
}

/*ARGSUSED*/
void PIiPwriterWait( fp, datatype, finalloc )
PIFILE *fp;
int    datatype, finalloc;
{
}

void PIiPreader( fp, buf, n, offset )
PIFILE *fp;
char   *buf;
int    n, offset;
{
int err; 

/* printf( "[%d] lroot = %d\n", PImytid, lroot ); */
/*    printf("!%d\n", offset); */
/* LSEEK( fp, offset, SEEK_SET, 0 ); */
err = READ( fp, buf+sizeof(long), n );
}

void PIiPreaderWait( fp, datatype )
PIFILE *fp;
int    datatype;
{
}

void PIiPwriterflush( fp )
PIFILE *fp;
{
}

