#ifndef PIO_BUF_SIZE
#include <fcntl.h>

/* Look for nonblocking operations */
#if defined(sun4)
#include <sys/asynch.h>
typedef aio_result_t PIaio_t;
#define PIO_HAS_NONBLOCKING_IO
#elif defined(intelnx)
typedef int PIaio_t;
#else
typedef int PIaio_t;
#endif

/* Read/write cache */
/* Eventually, this should include:
   int fd (file descriptor)
   int fpos (file position)
   anything else needed to make the cache read/write code independent of
   the PIFILE structure.  Then we can move this into tools.core/system,
   where it belongs.

   Also, we should consider allowing multiple read-ahead and write-behind
   buffers. 
 */
typedef struct {
    char    *rb, *wb;      /* Cache pointers */
    int     rp, rlen,      /* Read buffer and max length */
            wp, wlen,      /* Write buffer and max length */
            rsize;         /* Actual number of bytes in read buffer */
    int     rrloc, wrloc;  /* byte offset in file of beginning of read,
			      write cache */
    int     fpos;          /* file position relative to last IO op*/

    /* This part of the structure provides a place to mark that a 
       contiguous block of the file will be read/written, allowing
       more optimizations */
    int     spos, epos;    /* byte offsets in file; [spos:epos] will
			      be written */

    /* The remainder of this structure manages double buffering if
       nonblocking IO is used */
    PIaio_t rid, wid;      /* nonblocking IO hooks */
    char    *rb2, *wb2;    /* read-ahead and write-behind buffers */
    } PIFcache;

typedef struct {
    int     fd;
    long    fpos;
    int     mode, pmode;
    long    *sbuf;
    int     bsize;
    int     mtype;
    ProcSet *procset;
    int     rcnt;            /* Used to keep track of service requests */
    /* Here are the distributed writers and readers */
    void    (*diswriterSetup)(), (*disreaderSetup)();
    void    *disWctx, *disRctx;
    void    (*diswriter)(), (*diswriterwait)(), (*disflush)(),
            (*disreader)(), (*disreaderwait)();
    /* and the sequential seek, write, and read */
    int     (*lseek)(), (*write)(), (*read)(), (*flush)(), (*close)();
    PIFcache *c;
    } PIFILE;

/* This assumes a 0 origin for the relative values of start,end */
typedef struct {
    int mdim,               /* size of global array in this dimension */
        ndim,               /* size of local array in this dimension */
        start,              /* starting index for local piece */
        end;                /* ending index for local piece */
    int gstart, gend;       /* GLOBAL indices for start and end */
    } PIArrayPart;

extern PIFILE *PIFopen();
extern void   PIFclose(), PIFflush();
extern void   PIWriteCommon();
extern void   PIReadCommon();
extern void   PIWriteDistributedArray();
extern void   PIReadDistributedArray();

#define PIO_BUF_SIZE 1024

/* PIO Modes.  Associated with each of these should be a default reader/writer
   AS_SEQUENTIAL - same exact bits independent of the number of processors
   AS_PARALLEL   - any format; may only be readable by the same number of
                   processors on the same exact set of nodes (maximum 
		   performance)
   ANY_NUMBER    - (unimplemented) constraint on AS_PARALLEL; file readable
                   by any number of processors.  May force AS_SEQUENTIAL
		   depending on implementation.
 */
#define PIO_AS_SEQUENTIAL 0
#define PIO_AS_PARALLEL   0x1
#define PIO_ANY_NUMBER    0x2

/* Default cache sizes */
#ifndef PIF_RCACHE_SIZE
#define PIF_RCACHE_SIZE 4096
#define PIF_WCACHE_SIZE 4096
#endif

#endif
