/*
   This is an emulation library for intel nx, allowing nx programs to 
   use COMM as an interface.  However, a few things must be changed:
   
   The program initialization must use the PICall mechanism

   Also, not all routines are implemented; just the ones that we needed
 */   

#include "tools.h"
#include "comm/comm.h"
#include "system/system.h"
extern void PIdmax();

/*D
    NX2COMM - Using Chameleon to run programs written using Intel NX

    Description:
    Chameleon provides a backward compatibility mode for the NX
    message-passing routines.  By linking with the appropriate file, many
    NX programs can be run with few changes.

    The only change that you must make to the source code is to make the 
    program hostless and use the PICall interface.  Normally, this involves 
    only a few simple changes to main program.

    The other change is to insert the appropriate compatibilty interface
    module in the link line ahead of the Chameleon libraries.  The name of
    the interface file is ``nx2comm$(COMM).a'', where $(COMM) is the usual
    COMM variable (equal either to "", "p4", or "pvm").  For example,
    this makefile fragment links the NX program in pi.f with Chameleon
$
$   LDIR = /usr/local/tools.core/libs/libsg/$(ARCH)
$   $(FLINKER) -g -o pi pi.f $(LDIR)/nx2comm$(COMM).a \
$                 $(LDIR)/tools$(COMM).a $(LDIR)/tools.a \
$                 $(LDIR)/system.a $(CLIB)
$
    As usual, symbol CLIB is defined by the appropriate bmake include and ARCH 
    is the architecture.

    Notes:
    Not all (or even much) of NX is supported.  The supported routines 
    include csend, crecv, isend, irecv, msgwait, gsync, and gdsum.

    Rs6000 limitations:
    Because of the design of the C and Fortran compilers on the IBM rs6000's, 
    it is impossible to generate a single interface file that works with both
    languages (the design of Chameleon allows this, but NX's design does 
    not).  By default, the interface file FOR THE RS6000s ONLY is built ONLY
    for Fortran.  To build the C version INSTEAD, define BUILDNXC when
    compiling the interface file.  That is, use

$    xlc -Drs6000 -DBUILDNXC -D_POSIX_SOURCE -g -Dp4 \
$                           -I/usr/local/tools.core nx2comm.c

    to build the p4 version for C programs.
D*/

/* The intel versions manually remove the ``forcetype'' types and switch to 
   the forcetype code.  We should provide an option to assume ``intel source''
   that ALWAYS checks for forcetypes and removes the type information

   Another problem is that the NX names are monocase, preventing their use
   from both C and Fortran on IBM RS6000's.  Possible solutions:

   Define a Cname and a Fortran name; then use a switch to build EITHER a
   Fortran version or a C version of this library.  The Fortran code must
   NOT call the C versions of the emulation routines (have them call a common
   version in that case).

 */

/* 
   If neither FORTRANUNDERSCORE or FORTRANCAPS is defined, 
   we can NOT distinguish between the Fortran and C versions,
   so we have to decide whether we are generating the Fortran or the C
   version.  So far, only the rs6000 requires this version.  In this case,
   we look for the ADDITIONAL define: BUILDNXC .  In this case, 
   send0 etc will be Fortran versions; otherwise, they will be C versions.
*/
#if !defined(FORTRANUNDERSCORE) && !defined(FORTRANCAPS)
#ifndef BUILDNXC
#define NXFORTRANROUTINES
#else
#define NXCROUTINES
#endif /* BUILDNXC */
#else
#define NXCROUTINES
#define NXFORTRANROUTINES
#endif /* !defined(FORTRANUNDERSCORE) && !defined(FORTRANCAPS) */

#if defined(NXCROUTINES)
#define PCcsend      csend
#define PCcrecv      crecv 
#define PCisend      isend
#define PCirecv      irecv
#define PCmsgwait    msgwait
#define PCprobe      probe
#define PCrecvinfo   recvinfo
#define PCgsync      gsync
#define PCclock      clock
#define PCdclock     dclock
#define PCgdsum      gdsum
#define PCkillproc   killproc
#define PCmynode     mynode
#define PCnumnodes   numnodes
#define PCmypid      mypid
#define PCmclock     mclock
#define PCgdlow      gdlow
#define PCgdhigh     gdhigh
#define PCkillproc   killproc
#endif
#if defined(NXFORTRANROUTINES)
#define PFcsend     csend_
#define PFcrecv     crecv_
#define PFirecv     irecv_
#define PFisend     isend_
#define PFmsgwait   msgwait_
#define PFprobe     probe_
#define PFgsync     gsync_
#define PFclock     clock_
#define PFdclock    dclock_
#define PFgdsum     gdsum_
#define PFkillproc   killproc_
#define PFmynode     mynode_
#define PFnumnodes   numnodes_
#define PFmypid      mypid_
#define PFmclock     mclock_
#define PFgdlow      gdlow_
#define PFgdhigh     gdhigh_
#endif

/* mask off the top bits */
#define FORCEMASK (~0xc0000000)

void PCcsend(type, buf, len, node, pid )
int type, len, node, pid;
char *buf;
{
PIbsend(type&FORCEMASK,buf,len,node,MSG_OTHER);
}

void PCcrecv(type,buf,maxlen)
int  type, maxlen;
char *buf;
{
int tag;

tag = type;
if (tag >= 0) tag &= FORCEMASK;
PIbrecv(tag,buf,maxlen,MSG_OTHER);
}

int PCprobe(a)
int a;
{
return PInprobe(a);
}

void PCrecvinfo0(a,b,c) 
int *a, *b, *c;
{
*a = RECVLEN(); *b = RECVTYPE(); *c = RECVFROM(); 
}

void PCgsync()
{
PIgsync(ALLPROCS);
}

void PCbcast(a,b,c,d) 
char *a;
int  b, c, d;
{
GSCATTERSRC(a,b,d,ALLPROCS,MSG_OTHER);
}

double PCclock()
{
return SYGetElapsedTime();
}

void PCgdsum(a,b,c)
double *a, *c;
int  b;
{
PIgdsum( a, b, c, PSAllProcs ); 
}

double PCdclock()
{
return SYGetElapsedTime();
}

/* 
   Fortran versions.  At this point, if the Fortran versions are being
   defined, then PFsend0 -> send0_.  See if we need to further define
   these (say by changing them to all caps or remove the trailing underscore)
 */ 
#if defined(FORTRANCAPS)
#define csend_      CSEND
#define crecv_      CRECV
#define isend_      ISEND
#define irecv_      IRECV
#define msgwait_    MSGWAIT
#define probe_      PROBE
#define recvinfo_   RECVINFO
#define gsync_      GSYNC
#define clock_      CLOCK
#define dclock_     DCLOCK
#define gdsum_      GDSUM
#define killproc_   KILLPROC
#define mynode_     MYNODE
#define numnodes_   NUMNODES
#define mypid_      MYPID
#define mclock_     MCLOCK
#define gdlow_      GDLOW
#define gdhigh_     GDHIGH
#elif !defined(FORTRANUNDERSCORE)
#define csend_      csend
#define crecv_      crecv
#define isend_      isend
#define irecv_      irecv
#define msgwait_    msgwait
#define probe_      probe
#define gsync_      gsync
#define clock_      clock
#define dclock_     dclock
#define gdsum_      gdsum
#define killproc_   killproc
#define mynode_     mynode
#define numnodes_   numnodes
#define mypid_      mypid
#define mclock_     mclock
#define gdlow_      gdlow
#define gdhigh_     gdhigh
#endif

void PFcsend(type, buf, len, node, pid )
int *type, *len, *node, *pid;
char *buf;
{
PIbsend(*type&FORCEMASK,buf,*len,*node,MSG_OTHER);
}

void PFcrecv(a,b,c)
int  *b, *c;
char *a;
{
int tag = *c;
if (tag >= 0) tag &= FORCEMASK;
PIbrecv(tag,a,*b,MSG_OTHER);
}

int PFprobe(a)
int *a;
{
return PInprobe(*a);
}

void PFrecvinfo0(a,b,c) 
int *a, *b, *c;
{*a = RECVLEN(); *b = RECVTYPE(); *c = RECVFROM(); 
}

void PFgsync()
{
PIgsync(ALLPROCS);
}

void PFbcast(a,b,c,d) 
char *a;
int  *b, *c, *d;
{
GSCATTER(a,*b,(*d)==MYPROCID,ALLPROCS,MSG_OTHER);
}

double PFclock()
{
return SYGetElapsedTime();
}

void PFgdsum(a,b,c,d,e)
double *a, *c;
int  *b;
{
PCgdsum(a,*b,c);
}

double PFdclock()
{
return SYGetElapsedTime();
}

void PCkillproc( node, pid )
long node, pid;
{
if (node == -1) SYexitall( "killproc", 0 );
}
int PCmynode()
{
return PImytid;
}
int PCnumnodes()
{
return PInumtids;
}
int PCmypid()
{
return 0;
}
unsigned long PCmclock()
{
return (unsigned long)(SYGetCPUTime() * 1000.0);
}
void PCgdlow( val, n, buf )
double *val, *buf;
int    n;
{
GDMIN(val,n,buf,ALLPROCS);
}
void PCgdhigh( val, n, buf )    
double *val, *buf;
int    n;
{
GDMAX(val,n,buf,ALLPROCS);
}
void PFkillproc( node, pid )
long *node, *pid;
{
if (*node == -1) SYexitall( "killproc", 0 );
}
int PFmynode()
{
return MYPROCID;
}
int PFnumnodes()
{
return PInumtids;
}
int PFmypid()
{
return 0;
}
unsigned long PFmclock()
{
return (unsigned long)(SYGetCPUTime() * 1000.0);
}
void PFgdlow( val, n, buf )
double *val, *buf;
int    *n;
{
GDMIN(val,*n,buf,ALLPROCS);
}
void PFgdhigh( val, n, buf )    
double *val, *buf;
int    *n;
{
GDMAX(val,*n,buf,ALLPROCS);
}


/*
   To do irecv and isend, we need to see if "nonblocking" supported.
 */
#ifndef PI_NO_NSEND
int PCisend( type, buf, len, node, pid )
int type, len, node, pid;
char *buf;
{
int rc;
PInsend( type&FORCEMASK, buf, len, node, MSG_OTHER, rc )
return rc;
}
#else
/* don't save args since model is to send immediately */
#endif

#ifndef PI_NO_NRECV
int PCirecv( type, buf, len )
int type, len;
char *buf;
{
int rc;

if (type >= 0) type &= FORCEMASK;
PInrecv( type, buf, len, MSG_OTHER, rc );
return rc;
}
/* This version requies that wrecv ignores all but the last arg AND
   be independent of send/recv */
void PCmsgwait( rc )
int rc;
{
/* These parameters are for the eventual case where we handle the
   case of no nonblocking receive */
int type=0, len=0;
char *buf=0;
if (type >= 0) type &= FORCEMASK;
PIwrecv( type, buf, len, MSG_OTHER, rc );
}
#else
/* Must save the args since model is that actual recv happens at wait. */
#endif
