#ifndef lint
static char SCCSid[] = "@(#) ./comm/runpprog.c 12/01/93";
#endif

/*
    This file contains routines for "re-running" a program by passing it
    to another environment.  These routines

    Copy an argument list (to a specific position)
    Re-run a program (here is where we'll do things like try to handle
    file descriptors, deciding whether to run a program, etc
 */

#include "tools.h"

char *SYAllocAndCopy();

/* The format for the execv... routines is to null-terminate the argv list
   rather than pass the number of arguments.  Thus, we over-allocate Argv
   here and set the last element to null.  Also, if requote is true, 
   place "" around arguments that contain blanks.
 */   
void PICopyArgList( argc, argv, Argc, Argv, nargout, loc, requote )
int argc, *Argc, nargout, loc, requote;
char **argv, ***Argv;     
{
int  i;
char *p;
	
*Argv  = (char **) MALLOC( (nargout + 1) * sizeof(char*) );   CHKPTR(Argv);
*Argc = nargout;
for (i=0; i<=nargout; i++)
    (*Argv)[i] = (char *)0;
for (i=0; i + loc < nargout; i++) {
    if (strchr( argv[i+1], ' ' )) {
	p = (char *)MALLOC( strlen( argv[i+1] ) + 3 );
	sprintf( p, "\"%s\"", argv[i+1] );
	(*Argv)[i+loc] = p;
	}
    else {
	(*Argv)[i+loc] = (char *)MALLOC( strlen( argv[i+1] ) + 1 );
	strcpy( (*Argv)[i+loc], argv[i+1] );
	}
    }
}

/* 
  This routine looks at arguments.  If they contain a blank and are not
  quoted, they are replaced with a quoted version 
 */
void PIQuoteArgs( argc, argv )
int argc;
char **argv;
{
}

/*
   Compress an argument list into a single string.  Quote arguments that
   contain separators (blanks)
 */
void PICompressArgList( argc, argv, str, maxstr )
int argc, maxstr;
char **argv, *str;
{
char *p = str;
int  i;
for (i=0; i<argc; i++) {
    if (strchr( argv[i], ' ' )) {
	strcpy( p, "\"" );
	p ++;
	strcpy( p, argv[i] );
	p += strlen( argv[i] );
	strcpy( p, "\"" );
	p ++;
	strcpy( p, " " );
	p ++;
	}
    else {
	strcpy( p, argv[i] );
	p += strlen( argv[i] );
	strcpy( p, " " );
	p ++;
	}
    }
/* Remove the trailing blank */
if (p[-1] == ' ') p[-1] = 0;
}

/*+
    Run a program 
+*/
void PIRunProgram( name, argc, argv )
char *name, **argv;
int  argc;
{
#ifdef DOTEST
int i;
fprintf( stderr, "About to run " );
i = 0;
while (argv[i]) {
    fprintf( stderr, "\"%s\" ", argv[i] );
    i++;
    }
fprintf( stderr, "\n" );
#else
int rc;

rc = execvp( name, argv );
/* If we get here, there was a problem */
if (rc < 0) {
    perror( "Use -pispawned and the long startup form" );
    exit(1);
    }
#endif
}

/* We could make this a common interface, with the default behavior
   to just return */
#if (defined(euih) && defined(rs6000)) || DOTEST
/*+
    This is a special interface for EUI-H programs that resubmits the 
    job to the EUI-H job manager with the appropriate options and environment.
+*/
void PIiEUIHRun( argc, argv )
int  *argc;
char **argv;
{
int  np, Argc;
char snum[10], **Argv;
	
/* If I am the spawned job, just return */
if (SYArgHasName( argc, argv, 1, "-pispawned" )) 
    return;

/*
   The format of an EUI-H run is
   /usr/lpp/euih/eui/cb0eui -b argv[0] np -pispawned argv[1..]
 */
np = 1;
SYArgGetInt( argc, argv, 0, "-np", &np );

PICopyArgList( *argc, argv, &Argc, &Argv, *argc + 4, 5, 1 );

/* Eventually, we should get the euih run name and whether to use -b
   from an environment variable */
Argv[0] = SYAllocAndCopy( "/usr/lpp/euih/eui/cotb0" );
Argv[1] = SYAllocAndCopy( "-b" );
Argv[2] = SYAllocAndCopy( argv[0] );
sprintf( snum, "%d", np );
Argv[3] = SYAllocAndCopy( snum );
Argv[4] = SYAllocAndCopy( "-pispawned" );

/* Look for the hosts */
#ifdef FOO
if (SYArgHasName( argc, argv, 0, "-pihosts" )) {
    SYArgGetString( argc, argv, 0, "-pihosts", host_names, MAX_NAME_LEN );
    SYSetEnvironment( "EUI_HOSTS", host_names );
    /* putenv( name=value ); */
    }
#endif

/* Actually run program */
PIRunProgram( Argv[0], Argc, Argv );
}
#endif


#if (defined(intelparagon) && defined(FOO)) || DOTEST
/*+
    This is a special interface for paragon programs that resubmits the 
    job to the paragon job manager with the appropriate options and 
    environment.
+*/
void PIiParagonRun( argc, argv )
int  *argc;
char **argv;
{
int  np, Argc;
char snum[10], **Argv, *runstr;
	
/* If I am the spawned job, just return */
if (SYArgHasName( argc, argv, 1, "-pispawned" )) 
    return;

/*
   The format of an paragon run is 
   pexec "program and arguments" -sz np

   At least, for SOME paragons.  Caltech's paragon no longer accepts this
   form, so I have not finished/tested this code.
 */
np = 1;
SYArgGetInt( argc, argv, 0, "-np", &np );

runstr = (char *)MALLOC( 1025 );
PICompressArgList( *argc, argv, runstr, 1024 );
strcat( runstr, " -pispawned" );

Argv = (char **)MALLOC( 5 * sizeof(char **) );

Argv[0] = SYAllocAndCopy( "pexec" );
Argv[1] = runstr;
Argv[2] = SYAllocAndCopy( "-sz" );
sprintf( snum, "%d", np );
Argv[3] = SYAllocAndCopy( snum );
Argv[4] = 0;

/* Actually run program */
PIRunProgram( Argv[0], Argc, Argv );
}
#endif

char *SYAllocAndCopy( str )
char *str;
{
char *new;
new = (char *)MALLOC( strlen(str) + 1 );
strcpy( new, str );
return new;
}	

/*
    For EUI jobs, change prog -np n ... to prog -procs n -np n ...
    For Paragon jobs, change prog -np n ... to pexec ????

    For other systems, the jobs run on a different front end, and hence can
    not just re-exec themselves.  In this case, get Harrison's program.
 */

#ifdef DOTEST
main( argc, argv )
int  argc;
char **argv;
{
int Argc = argc;
PIiParagonRun( &Argc, argv );
PIiEUIHRun( &Argc, argv );
}
#endif
