#ifndef lint
static char SCCSid[] = "@(#) ./system/state.c 12/01/93";
#endif

#include "tools.h"
#include "system/state.h"     /*I "system/state.h" I*/

/* 
   This file contains routines for a simple state tracking facility.
 */

#define MAX_STACK 25
static SYusc_time_t timestk[MAX_STACK+1];
static int          sp = 0;           /* points to next free one */

/*@
  SYStateInit - initialize a state

  Input Parameter:
. state - state to initialize (all fields set to zero)
@*/
void SYStateInit( state )
SYstate *state;
{
state->time    = 0.0;
state->count   = 0;
state->data    = 0;
state->routine = 0;
state->rcontext= 0;
state->maxtime = 0.0;
}

/*@
   SYStatePush - push a state 

    Input parameters:
.   state - state to use
.   data  - data to add into the states data field
@*/  
void SYStatePush( state, data )
SYstate *state;
int     data;
{
state->data += data;
if (sp < MAX_STACK) {
    SYusc_clock( timestk + sp );
    if (state->routine) (*state->routine)( state, 1, data, 0.0 );
    sp ++;
    }
}

/*@
    SYStatePop - pop a state

    Input parameters:
.   state - state to use
.   data  - data to add into the states data field
@*/
void SYStatePop( state, data, file, line )
SYstate *state;
int     data;
char    *file;
int     line;
{
SYusc_time_t t;
double       tval;

SYusc_clock( &t );
state->count++;
state->data += data;
if (sp) sp--;
tval         = SYuscDiff( timestk + sp, &t );
state->time += tval;
/* Keep track of where the "winner" is */
if (tval > state->maxtime) {
    state->maxtime = tval;
    state->line    = line;
    strncpy( state->file, file, SYSTATEFILELEN );
    }
if (state->routine) (*state->routine)( state, 0, data, tval );
}

